using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using Microsoft.DirectX.Direct3D;

namespace BenLincoln.TheLostWorlds.CDTextures
{
    public class SoulReaverDreamcastTextureFile : SoulReaverMonolithicTextureFile
    {
        // Methods
        public SoulReaverDreamcastTextureFile(string path)
            : base(path)
        {
            base._FileType = TextureFileType.SoulReaverDreamcast;
            base._FileTypeName = "Soul Reaver (Dreamcast) Vector-Quantized Monolithic Texture File";
            base._HeaderLength = 0x4800;
            base._TextureLength = 0x4800;
            base._TextureCount = this._GetTextureCount();
        }

        protected override Texture _GetTexture(Device device, int index)
        {
            return Texture.FromBitmap(device, this._GetTextureAsBitmap(index), Usage.None, Pool.Managed);
        }

        protected override Bitmap _GetTextureAsBitmap(int index)
        {
            Bitmap bitmap;
            int num = base._HeaderLength + (index * base._TextureLength);
            if (num > (base._FileInfo.Length - base._TextureLength))
            {
                throw new TextureFileException("An index was specified which resulted in an offset greater than the maximum allowed value.");
            }
            try
            {
                FileStream input = new FileStream(base._FilePath, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(input);
                input.Seek((long)num, SeekOrigin.Begin);
                ushort[] codeBook = new ushort[0x400];
                for (int i = 0; i < 0x400; i++)
                {
                    codeBook[i] = reader.ReadUInt16();
                }
                byte[] imageData = new byte[0x4000];
                for (int j = 0; j < imageData.GetUpperBound(0); j++)
                {
                    imageData[j] = reader.ReadByte();
                }
                reader.Close();
                input.Close();
                bitmap = this.decodeTwiddledVQ(imageData, codeBook, 0x100, 0x100);
            }
            catch (Exception exception)
            {
                throw new TextureFileException("Error reading the specified texture.", exception);
            }
            return bitmap;
        }

        protected Bitmap decodeTwiddledVQ(byte[] imageData, ushort[] codeBook, int width, int height)
        {
            int num4 = 15;
            int num5 = 10;
            int num6 = 5;
            int num7 = 0;
            Bitmap bitmap = new Bitmap(width, height);
            int[] numArray = this.initTwiddleTable();
            Color[] colorArray = new Color[0x400];
            int index = 0;
            while (index < 0x400)
            {
                ushort num8 = codeBook[index];
                int alpha = (num8 >> num4) & 1;
                if (alpha > 0)
                {
                    alpha = 0xff;
                }
                colorArray[index] = Color.FromArgb(alpha, ((num8 >> (num5 & 0x1f)) & 0x1f) << 3, ((num8 >> (num6 & 0x1f)) & 0x1f) << 3, ((num8 >> (num7 & 0x1f)) & 0x1f) << 3);
                index++;
            }
            for (int i = 0; i < height; i += 2)
            {
                for (index = 0; index < width; index += 2)
                {
                    int num9 = imageData[numArray[i / 2] | (numArray[index / 2] << 1)];
                    int num10 = num9 * 4;
                    Color color = colorArray[num10];
                    Color color2 = colorArray[num10 + 1];
                    Color color3 = colorArray[num10 + 2];
                    Color color4 = colorArray[num10 + 3];
                    bitmap.SetPixel(index, i, color);
                    bitmap.SetPixel(index + 1, i, color3);
                    bitmap.SetPixel(index, i + 1, color2);
                    bitmap.SetPixel(index + 1, i + 1, color4);
                }
            }
            return bitmap;
        }

        public override MemoryStream GetDataAsStream(int index)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        protected int[] initTwiddleTable()
        {
            int num;
            int[] numArray = new int[0x400];
            for (num = 0; num < 0x20; num++)
            {
                numArray[num] = ((((num & 1) | ((num & 2) << 1)) | ((num & 4) << 2)) | ((num & 8) << 3)) | ((num & 0x10) << 4);
            }
            for (num = 0x20; num < 0x400; num++)
            {
                numArray[num] = numArray[num & 0x1f] | (numArray[(num >> 5) & 0x1f] << 10);
            }
            return numArray;
        }
    }
}
