using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace ModelEx3
{
    public interface ModelExFile
    {
        String FileName { get; }
    }

    class ModelExFileList<T> where T : ModelExFile
    {
        protected enum Action
        {
            Add = 0,
            Find = 1,
            Remove = 2,
        }

        protected List<WeakReference> m_xReferences;
        public int Count { get { return m_xReferences.Count; } }

        public T this[int iIndex]
        {
            get
            {
                if (m_xReferences[iIndex].Target != null && m_xReferences[iIndex].IsAlive)
                {
                    return (T)m_xReferences[iIndex].Target;
                }
                return default(T);
            }
        }

        public ModelExFileList()
        {
            m_xReferences = new List<WeakReference>();
        }

        public T Add(T xItem)
        {
            return Find(xItem, Action.Add);
        }

        public void Remove(T xItem)
        {
            Find(xItem, Action.Remove);
        }

        public void Clear()
        {
            m_xReferences.Clear();
        }

        public T Find(T xItem)
        {
            return Find(xItem, Action.Find);
        }

        protected T Find(T xItem, Action eAction)
        {
            T xResult = default(T);
            List<WeakReference> xDeadReferences = new List<WeakReference>();
            foreach (WeakReference xReference in m_xReferences)
            {
                if (xReference != null && xReference.IsAlive)
                {
                    T xExistingItem = (T)xReference.Target;
                    if (xItem.FileName == xExistingItem.FileName)
                    {
                        if (eAction == Action.Find)
                        {
                            xResult = xExistingItem;
                        }
                        if (eAction == Action.Remove)
                        {
                            xDeadReferences.Add(xReference);
                        }
                    }
                }
                else
                {
                    xDeadReferences.Add(xReference);
                }
            }

            if (eAction == Action.Add)
            {
                m_xReferences.Add(new WeakReference(xItem));
            }

            foreach (WeakReference xDeadReference in xDeadReferences)
            {
                m_xReferences.Remove(xDeadReference);
            }
            xDeadReferences.Clear();

            return xResult;
        }

        public T Find(String strFileName)
        {
            foreach (WeakReference xReference in m_xReferences)
            {
                if (xReference != null && xReference.IsAlive)
                {
                    T xExistingItem = (T)xReference.Target;
                    if (xExistingItem.FileName == strFileName)
                    {
                        return xExistingItem;
                    }
                }
            }
            return default(T);
        }

        public T[] ToArray()
        {
            List<T> xList = new List<T>();
            foreach (WeakReference xReference in m_xReferences)
            {
                if (xReference != null && xReference.IsAlive)
                {
                    xList.Add((T)xReference.Target);
                }
            }
            return xList.ToArray();
        }
    }
}
