using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    class ModelExMouseController : ModelExArcBall
    {
        public enum ControlMode
        {
            XYZ = 0,
            XY = 1
        };

        Vector2     m_xScreenPosition;
        ControlMode m_eMode;

        public ControlMode Mode
        {
            get { return m_eMode; }
            set { m_eMode = value; }
        }

        public ModelExMouseController(ref Control xControl, bool bInversedControls)
            : base(ref xControl, bInversedControls)
        {
            m_xScreenPosition = new Vector2(0.0f, 0.0f);
            m_eMode = ControlMode.XYZ;
        }

        protected override void HandleMove(int iX, int iY, Quaternion xReference)
        {
            if (m_eMode == ControlMode.XY)
            {
                if (m_bBeginMove)
                {
                    m_xScreenPosition = new Vector2(iX, iY);
                }

                float fYaw = (float) Math.PI * (iX - m_xScreenPosition.X) / m_xControl.Width;
                float fPitch = (float) Math.PI * (iY - m_xScreenPosition.Y) / m_xControl.Height;

                Quaternion xRotateYaw = Quaternion.RotationAxis(new Vector3(0.0f, 1.0f, 0.0f), fYaw);
                m_xCurrentRotation *= xRotateYaw;
                m_xCurrentRotation.Normalize();

                Quaternion xRotatePitch = Quaternion.RotationAxis(new Vector3(1.0f, 0.0f, 0.0f), fPitch);
                m_xCurrentRotation = xRotatePitch * m_xCurrentRotation;
                m_xCurrentRotation.Normalize();

                m_xScreenPosition = new Vector2(iX, iY);
            }
            else
            {
                base.HandleMove(iX, iY, xReference);
            }
        }
    }
}
