using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

namespace TrackBall
{
    public class Trackball
    {
        protected int m_Width, m_Height;
        protected Vector3 m_StartProjection, m_CurrentProjection;
        protected Quaternion m_StartRotation, m_CurrentRotation;

        public Trackball()
        {
            m_Width = 0;
            m_Height = 0;
            m_StartProjection = Vector3.Zero;
            m_CurrentProjection = Vector3.Zero;
            m_StartRotation = Quaternion.Identity;
            m_CurrentRotation = Quaternion.Identity;
        }

        public void Resize(int width, int height)
        {
            m_Width = width;
            m_Height = height;
        }

        public Vector2 ScreenToNDC(int mouseX, int mouseY)
        {
            float x = 0;
            float y = 0;

            mouseY = m_Height - mouseY;

            x = (2.0f * mouseX - m_Width) / (float)m_Width;
            y = (2.0f * mouseY - m_Height) / (float)m_Height;

            return new Vector2(x, y);
        }

        // Project the mouse coordinates onto the sphere
        public Vector3 ProjectToSphere(Vector2 ndc)
        {
            Vector3 theVector = new Vector3(ndc.X, ndc.Y, 0);
            float theLength = ndc.Length();
            // If the mouse coordinate lies outside the sphere
            // choose the closest point on the sphere by 
            // setting z to zero and renomalizing
            if (theLength >= 1.0f)
            {
                theVector.X /= (float)Math.Sqrt(theLength);
                theVector.Y /= (float)Math.Sqrt(theLength);
            }
            else
            {
                theVector.Z = 1.0f - theLength;
            }

            return theVector;
        }

        public Quaternion GetQuaternionFromProjections(Vector3 vFrom, Vector3 vTo)
        {
            Vector3 vAxis = Vector3.Cross(vFrom, vTo);
            return new Quaternion(vAxis.X, vAxis.Y, vAxis.Z, Vector3.Dot(vFrom, vTo));
        }

        public Quaternion GetQuaternion()
        {
            Quaternion theRotation = new Quaternion(m_CurrentRotation.X,
                                                    m_CurrentRotation.Y,
                                                    m_CurrentRotation.Z,
                                                    m_CurrentRotation.W);
            theRotation.Normalize();
            return theRotation;
        }

        public Matrix GetMatrix()
        {
            Quaternion qRotation = GetQuaternion();
            Matrix theMatrix = Matrix.CreateFromQuaternion(qRotation);
            return theMatrix;
        }

        public void Begin(int x, int y)
        {
            Vector2 ndc = ScreenToNDC(x, y);
            m_StartProjection = ProjectToSphere(ndc);
            m_StartRotation = m_CurrentRotation;
        }

        public void Update(int x, int y)
        {
            Vector2 ndc = ScreenToNDC(x, y);
            m_CurrentProjection = ProjectToSphere(ndc);
            m_CurrentRotation = GetQuaternionFromProjections(m_StartProjection, m_CurrentProjection) * m_StartRotation;
        }
    }
}

